#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/kthread.h>
#include <linux/delay.h> /* mdelay */

#include "portal.h"  
#include "dmaManager.h"
#include "GeneratedTypes.h" 

#include "bdbm_drv.h"
#include "params.h"
#include "debug.h"
#include "dm_bluedbm.h"


#define MAX_INDARRAY 4
DECLARE_COMPLETION(worker_completion);
static DmaManagerPrivate priv;
static PortalInternal intarr[MAX_INDARRAY];
static sem_t test_sem;

ssize_t xbsv_kernel_read (struct file *f, char __user *arg, size_t len, loff_t *data);
ssize_t xbsv_kernel_write (struct file *f, const char __user *arg, size_t len, loff_t *data);
struct semaphore bsim_start;
int main_program_finished = 0;
int bsim_relay_running;
int nandsim_init = 0;

static struct file_operations pa_fops = {
	.owner = THIS_MODULE,
};

static struct miscdevice miscdev = {
	.minor = MISC_DYNAMIC_MINOR,  // Must be < 256!
	.name = "xbsvtest",
	.fops = &pa_fops,
};

uint32_t erase_ret = 0;


/* call-back functions */
void FlashIndicationWrapperreadDone_cb (
	struct PortalInternal *p, 
	const uint32_t tag)
{
	/*bdbm_msg ("cb: FlashIndicationWrapperreadDone_cb (tag = %x)", tag);*/
	sem_post (&test_sem);
}

void FlashIndicationWrapperwriteDone_cb (
	struct PortalInternal *p, 
	const uint32_t tag)
{
	/*bdbm_msg ( "cb: FlashIndicationWrapperwriteDone_cb (tag = %x)", tag);*/
	sem_post (&test_sem);
}

void FlashIndicationWrappereraseDone_cb (
	struct PortalInternal *p, 
	const uint32_t tag, 
	const uint32_t status)
{
	/*bdbm_msg ( "cb: FlashIndicationWrappereraseDone_cb (tag = %x status = %x)", tag, status);*/
	if (status != 0) {
		bdbm_msg ("cb: oops! bad block");
		erase_ret = 1;
	}
	sem_post (&test_sem);
}

void FlashIndicationWrapperdebugDumpResp_cb (
	struct PortalInternal *p, 
	const uint32_t debug0, 
	const uint32_t debug1, 
	const uint32_t debug2, 
	const uint32_t debug3 )
{
	bdbm_msg ( "cb: FlashIndicationWrapperdebugDumpResp_cb");
	sem_post (&test_sem);
}

void MMUConfigIndicationWrapperidResponse_cb (
	struct PortalInternal *p, 
	const uint32_t sglId) 
{
	bdbm_msg ("cb: MMUConfigIndicationWrapperidResponse_cb");
	priv.sglId = sglId;
	sem_post (&priv.sglIdSem);
}

void MMUConfigIndicationWrapperconfigResp_cb (
	struct PortalInternal *p, 
	const uint32_t pointer)
{
	bdbm_msg ("cb: MMUConfigIndicationWrapperconfigResp_cb(physAddr=%x)", pointer);
	sem_post (&priv.confSem);
}

void MMUConfigIndicationWrappererror_cb (
	struct PortalInternal *p, 
	const uint32_t code, 
	const uint32_t pointer, 
	const uint64_t offset, 
	const uint64_t extra)
{
	bdbm_msg ("cb: MMUConfigIndicationWrappererror_cb");
}

void DmaDebugIndicationWrapperaddrResponse_cb (
	struct PortalInternal *p, 
	const uint64_t physAddr)
{
	bdbm_msg ("cb: DmaDebugIndicationWrapperaddrResponse_cb");
}

void DmaDebugIndicationWrapperreportStateDbg_cb (
	struct PortalInternal *p, 
	const DmaDbgRec rec )
{
	bdbm_msg ("cb: DmaDebugIndicationWrapperreportStateDbg_cb");
	sem_post (&priv.dbgSem);
}

void DmaDebugIndicationWrapperreportMemoryTraffic_cb (
	struct PortalInternal *p, 
	const uint64_t words)
{
	bdbm_msg ("cb: DmaDebugIndicationWrapperreportMemoryTraffic_cb");
	priv.mtCnt = words;
	sem_post (&priv.mtSem);
}

void DmaDebugIndicationWrappererror_cb (
	struct PortalInternal *p, 
	const uint32_t code, 
	const uint32_t pointer, 
	const uint64_t offset, 
	const uint64_t extra) 
{
	bdbm_msg ("cb: DmaDebugIndicationWrappererror_cb");
}

void manual_event(void)
{
	int i;
	for (i = 0; i < MAX_INDARRAY; i++) {
		PortalInternal *instance = &intarr[i];
		volatile uint32_t *map_base = instance->map_base;
		uint32_t queue_status;
		while ((queue_status= READL(instance, &map_base[IND_REG_QUEUE_STATUS]))) {
			uint32_t int_src, int_en, ind_count;
			int_src = READL (instance, &map_base[IND_REG_INTERRUPT_FLAG]);
			int_en = READL (instance, &map_base[IND_REG_INTERRUPT_MASK]);
			ind_count = READL (instance, &map_base[IND_REG_INTERRUPT_COUNT]);
			instance->handler (instance, queue_status-1);
		}
	}
}

int pthread_worker_fn (void* arg) 
{
	while (1) {
		manual_event ();
		yield ();
		/*msleep (0);*/
		if (kthread_should_stop ()) {
			bdbm_msg ("pthread_worker_fn ends");
			break;
		}
	}
	complete (&worker_completion);
	return 0;
}

/**
 * data structures for dm_bluedbm
 */
#define FLASH_PAGE_SIZE 8192
#define NUM_TAGS 128

int srcAlloc;
int dstAlloc;
size_t dstAlloc_sz = FLASH_PAGE_SIZE * NUM_TAGS *sizeof(uint8_t);
size_t srcAlloc_sz = FLASH_PAGE_SIZE * NUM_TAGS *sizeof(uint8_t);
uint32_t ref_dstAlloc; 
uint32_t ref_srcAlloc; 
uint32_t* dstBuffer;
uint32_t* srcBuffer;
uint32_t* readBuffers[NUM_TAGS];
uint32_t* writeBuffers[NUM_TAGS];
bool dstBufBusy[NUM_TAGS]; 
bool srcBufBusy[NUM_TAGS]; 
bool eraseTagBusy[NUM_TAGS]; 

struct task_struct *worker = NULL;

struct bdbm_dm_inf_t _dm_bluedbm_inf = {
	.ptr_private = NULL,
	.probe = dm_bluedbm_probe,
	.open = dm_bluedbm_open,
	.close = dm_bluedbm_close,
	.make_req = dm_bluedbm_make_req,
	.end_req = dm_bluedbm_end_req,
	.load = NULL,
	.store = NULL,
};

struct dm_bluedbm_private {
};


uint32_t dm_bluedbm_probe (struct bdbm_drv_info* bdi)
{
	bdbm_msg ("dm_bluedbm_probe (minor %d) - begin", miscdev.minor);

	misc_register (&miscdev);
	if ((worker = kthread_create (pthread_worker_fn, NULL, "pthread_worker_fn")) == NULL) {
		bdbm_error ("kthread_create failed");
		goto fail;
	}
	sema_init (&bsim_start, 0);
	nandsim_init = 1;

	bdbm_msg ("dm_bluedbm_probe (minor %d) - end", miscdev.minor);

	return 0;

fail:
	bdbm_error ("dm_bluedbm_probe failed!");
	misc_deregister (&miscdev);
	return 1;
}

uint32_t dm_bluedbm_open (struct bdbm_drv_info* bdi)
{
	int i = 0, t = 0;

	bdbm_msg ("dm_bluedbm_open - begin");

	/* create portals */
	init_portal_internal (&intarr[2], IfcNames_HostMMUConfigRequest, NULL); // fpga3
	init_portal_internal (&intarr[0], IfcNames_HostMMUConfigIndication, MMUConfigIndicationWrapper_handleMessage); // fpga1
	init_portal_internal (&intarr[3], IfcNames_FlashRequest, NULL); // fpga4
	init_portal_internal (&intarr[1], IfcNames_FlashIndication, FlashIndicationWrapper_handleMessage); // fpga2
	DmaManager_init (&priv, NULL, &intarr[2]);
	sem_init (&test_sem, 0, 0);

	/* create and run a thread for message handling */
	wake_up_process (worker);

	/* allocate memory */
	srcAlloc = portalAlloc (srcAlloc_sz);
	dstAlloc = portalAlloc (dstAlloc_sz);
	srcBuffer = (uint32_t*)portalMmap (srcAlloc, srcAlloc_sz);
	dstBuffer = (uint32_t*)portalMmap (dstAlloc, dstAlloc_sz);

	bdbm_msg ("dstAlloc = %x", dstAlloc); 
	bdbm_msg ("srcAlloc = %x", srcAlloc); 
	
	portalDCacheFlushInval(dstAlloc, dstAlloc_sz, dstBuffer);
	portalDCacheFlushInval(srcAlloc, srcAlloc_sz, srcBuffer);
	ref_dstAlloc = DmaManager_reference (&priv, dstAlloc);
	ref_srcAlloc = DmaManager_reference (&priv, srcAlloc);

	/* assign & initialize memory */
	for (t = 0; t < NUM_TAGS; t++) {
		int byteOffset = t * FLASH_PAGE_SIZE;
		dstBufBusy[t] = false;
		srcBufBusy[t] = false;
		FlashRequestProxy_addDmaWriteRefs (&intarr[3], ref_dstAlloc, byteOffset, t);
		FlashRequestProxy_addDmaReadRefs (&intarr[3], ref_srcAlloc, byteOffset, t);
		readBuffers[t] = dstBuffer + byteOffset/sizeof(uint32_t);
		writeBuffers[t] = srcBuffer + byteOffset/sizeof(uint32_t);
	}

	for (t = 0; t < NUM_TAGS; t++) {
		for (i = 0; i < PAGE_SIZE/sizeof(uint32_t); i++ ) {
			readBuffers[t][i] = 0;
			writeBuffers[t][i] = 0;
			/*writeBuffers[t][i] = (t<<16)+i;*/
		}
	}

	/* init a device */
	FlashRequestProxy_start (&intarr[3], 0);
	FlashRequestProxy_setDebugVals (&intarr[3], 0, 0);
	FlashRequestProxy_debugDumpReq (&intarr[3], 0);
	sem_wait (&test_sem);
	FlashRequestProxy_debugDumpReq (&intarr[3], 0);
	sem_wait (&test_sem);

	bdbm_msg ("dm_bluedbm_open - end");

	return 0;
}

void dm_bluedbm_close (struct bdbm_drv_info* bdi)
{
	bdbm_msg ("dm_bluedbm_close - begin");

	if (!nandsim_init) {
		bdbm_error ("dm_bluedbm is not initialized yet");
		return;
	}
	main_program_finished = 1;

	bdbm_msg ("wait for a worker thread to finish");
	kthread_stop (worker);
	wait_for_completion_timeout (&worker_completion, msecs_to_jiffies(2000));

	misc_deregister (&miscdev);

	bdbm_msg ("dm_bluedbm_close - end");
}

void __dm_bluedbm_cp_dma_to_bio (
	struct bdbm_drv_info* bdi,
	struct bdbm_llm_req_t* r)
{
	struct nand_params* np = BDBM_GET_NAND_PARAMS (bdi);
	uint8_t* ptr_dma_addr = NULL;
	uint32_t nr_pages = 0;
	uint32_t loop = 0;

	nr_pages = np->page_main_size / KERNEL_PAGE_SIZE;
	ptr_dma_addr = (uint8_t*)readBuffers[0];

	/* copy the main page data to a buffer */
	for (loop = 0; loop < nr_pages; loop++) {
		if (r->req_type == REQTYPE_RMW_READ && 
			r->kpg_flags[loop] == MEMFLAG_KMAP_PAGE) {
			continue;
		}

		bdbm_memcpy (
			r->pptr_kpgs[loop],
			ptr_dma_addr + KERNEL_PAGE_SIZE * loop, 
			KERNEL_PAGE_SIZE
		);
	}

	/* copy the OOB data to a buffer */
	if (r->req_type != REQTYPE_RMW_READ && 
		r->ptr_oob != NULL &&
		np->page_oob_size != 0) {
		bdbm_memcpy (
			r->ptr_oob, 
			ptr_dma_addr + np->page_main_size,
			np->page_oob_size
		);
	}
}

void __dm_bluedbm_cp_bio_to_dma (
	struct bdbm_drv_info* bdi,
	struct bdbm_llm_req_t* r)
{
	struct nand_params* np = BDBM_GET_NAND_PARAMS (bdi);
 	uint8_t* ptr_dma_addr = NULL;
	uint32_t nr_pages = 0;
	uint32_t loop = 0;

	nr_pages = np->page_main_size / KERNEL_PAGE_SIZE;
	ptr_dma_addr = (uint8_t*)writeBuffers[0];

	/* copy the main page data to a buffer */
	for (loop = 0; loop < nr_pages; loop++) {
		bdbm_memcpy (
			ptr_dma_addr + KERNEL_PAGE_SIZE * loop, 
			r->pptr_kpgs[loop], 
			KERNEL_PAGE_SIZE
		);
	}

	/* copy the OOB data to a buffer */
	if (np->page_oob_size != 0 && r->ptr_oob != NULL) {
		bdbm_memcpy (
			ptr_dma_addr + np->page_main_size,
			r->ptr_oob,
			np->page_oob_size
		);
	}
}

uint32_t dm_bluedbm_make_req (
	struct bdbm_drv_info* bdi, 
	struct bdbm_llm_req_t* ptr_llm_req)
{
	struct bdbm_phyaddr_t* phyaddr = ptr_llm_req->phyaddr;

	/*bdbm_msg ("dm_bluedbm_make_req - begin");*/

	/* send a request to bluesime according to its type */
	switch (ptr_llm_req->req_type) {
	case REQTYPE_WRITE:
	case REQTYPE_RMW_WRITE:
	case REQTYPE_GC_WRITE:
		__dm_bluedbm_cp_bio_to_dma (bdi, ptr_llm_req);
		FlashRequestProxy_writePage (
			&intarr[3], 
			phyaddr->channel_no, 
			phyaddr->chip_no, 
			phyaddr->block_no, 
			phyaddr->page_no, 
			0);
		sem_wait(&test_sem);

		FlashRequestProxy_readPage (
			&intarr[3], 
			phyaddr->channel_no, 
			phyaddr->chip_no, 
			phyaddr->block_no, 
			phyaddr->page_no, 
			0);
		sem_wait(&test_sem);
		/*__dm_bluedbm_cp_dma_to_bio (bdi, ptr_llm_req);*/

		if (memcmp (writeBuffers[0], readBuffers[0], KERNEL_PAGE_SIZE*2) != 0) {
			bdbm_msg ("%x %x %x %x %x %x %x %x", 
				writeBuffers[0][0],
				writeBuffers[0][1],
				writeBuffers[0][2],
				writeBuffers[0][3],
				writeBuffers[0][4],
				writeBuffers[0][5],
				writeBuffers[0][6],
				writeBuffers[0][7]);

			bdbm_msg ("%x %x %x %x %x %x %x %x", 
				readBuffers[0][0],
				readBuffers[0][1],
				readBuffers[0][2],
				readBuffers[0][3],
				readBuffers[0][4],
				readBuffers[0][5],
				readBuffers[0][6],
				readBuffers[0][7]);
		}
		break;

	case REQTYPE_READ:
	case REQTYPE_RMW_READ:
	case REQTYPE_GC_READ:
		FlashRequestProxy_readPage (
			&intarr[3], 
			phyaddr->channel_no, 
			phyaddr->chip_no, 
			phyaddr->block_no, 
			phyaddr->page_no, 
			0);
		sem_wait(&test_sem);
		__dm_bluedbm_cp_dma_to_bio (bdi, ptr_llm_req);
		break;

	case REQTYPE_GC_ERASE:
		erase_ret = 0;
		FlashRequestProxy_eraseBlock (&intarr[3], 
			phyaddr->channel_no, 
			phyaddr->chip_no, 
			phyaddr->block_no, 
			0);
		sem_wait(&test_sem);
		if (erase_ret != 0) {
			ptr_llm_req->ret = 1;
		}
		break;

	case REQTYPE_READ_DUMMY:
	default:
		break;
	}

	/*bdbm_msg ("dm_bluedbm_make_req - end");*/

	bdi->ptr_dm_inf->end_req (bdi, ptr_llm_req);
	return 0;
}

void dm_bluedbm_end_req (
	struct bdbm_drv_info* bdi, 
	struct bdbm_llm_req_t* ptr_llm_req)
{
	bdi->ptr_llm_inf->end_req (bdi, ptr_llm_req);
}

