#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/kthread.h>
#include <linux/delay.h> /* mdelay */

#include "debug.h"
#include "portal.h"  
#include "dmaManager.h"
#include "GeneratedTypes.h" 


static int nandsim_init = 0;

/* for sock_utils.c */
ssize_t xbsv_kernel_read (struct file *f, char __user *arg, size_t len, loff_t *data);
ssize_t xbsv_kernel_write (struct file *f, const char __user *arg, size_t len, loff_t *data);
struct semaphore bsim_start;
int main_program_finished = 0;
int bsim_relay_running;

static struct file_operations pa_fops = {
	.owner = THIS_MODULE,
	.read = xbsv_kernel_read,
	.write = xbsv_kernel_write,
};

static struct miscdevice miscdev = {
	.minor = MISC_DYNAMIC_MINOR,  // Must be < 256!
	.name = "xbsvtest",
	.fops = &pa_fops,
};

/* for kernel threads */
DECLARE_COMPLETION(main_completion);
DECLARE_COMPLETION(worker_completion);
static struct task_struct *kid_main = NULL;
static struct task_struct *kid_worker = NULL;
static int trace_memory;// = 1;

#define MAX_INDARRAY 4
static PortalInternal intarr[MAX_INDARRAY];

static sem_t test_sem;
/*static int burstLen = 16;*/
#ifdef BSIM
#define numWords 0x1240000/4 // make sure to allocate at least one entry of each size
#else
#define numWords 0x124000/4
#endif
/*static long test_sz  = numWords*sizeof(unsigned int);*/
static long alloc_sz = numWords*sizeof(unsigned int);
static DmaManagerPrivate priv;
size_t numBytes = 1 << 12;


/* call-back functions */
void NandSimIndicationWrappereraseDone_cb (  struct PortalInternal *p, const uint32_t tag )
{
         PORTAL_PRINTF( "NandSim_eraseDone(tag = %x)\n", tag);
         sem_post(&test_sem);
}
void NandSimIndicationWrapperwriteDone_cb (  struct PortalInternal *p, const uint32_t tag )
{
         PORTAL_PRINTF( "NandSim_writeDone(tag = %x)\n", tag);
         sem_post(&test_sem);
}
void NandSimIndicationWrapperreadDone_cb (  struct PortalInternal *p, const uint32_t tag )
{
         PORTAL_PRINTF( "NandSim_readDone(tag = %x)\n", tag);
         sem_post(&test_sem);
}
void DmaIndicationWrapperconfigResp_cb (  struct PortalInternal *p, const uint32_t pointer )
{
        PORTAL_PRINTF("DmaIndication_configResp(physAddr=%x)\n", pointer);
        sem_post(&priv.confSem);
}
void DmaIndicationWrapperaddrResponse_cb (  struct PortalInternal *p, const uint64_t physAddr )
{
        PORTAL_PRINTF("DmaIndication_addrResponse(physAddr=%"PRIx64")\n", physAddr);
}
void DmaIndicationWrapperreportStateDbg_cb (  struct PortalInternal *p, const DmaDbgRec rec )
{
        PORTAL_PRINTF("reportStateDbg: {x:%08x y:%08x z:%08x w:%08x}\n", rec.x,rec.y,rec.z,rec.w);
        sem_post(&priv.dbgSem);
}
void DmaIndicationWrapperreportMemoryTraffic_cb (  struct PortalInternal *p, const uint64_t words )
{
        //PORTAL_PRINTF("reportMemoryTraffic: words=%"PRIx64"\n", words);
        priv.mtCnt = words;
        sem_post(&priv.mtSem);
}
void DmaIndicationWrapperdmaError_cb (  struct PortalInternal *p, const uint32_t code, const uint32_t pointer, const uint64_t offset, const uint64_t extra ) 
{
        PORTAL_PRINTF("DmaIndication::dmaError(code=%x, pointer=%x, offset=%"PRIx64" extra=%"PRIx64"\n", code, pointer, offset, extra);
}

void manual_event(void)
{
    int i;
    for (i = 0; i < MAX_INDARRAY; i++) {
      PortalInternal *instance = &intarr[i];
      volatile unsigned int *map_base = instance->map_base;
      unsigned int queue_status;
	  /*PORTAL_PRINTF ("[%d]\n", i);*/
      while ((queue_status= READL(instance, &map_base[IND_REG_QUEUE_STATUS]))) {
        unsigned int int_src = READL(instance, &map_base[IND_REG_INTERRUPT_FLAG]);
        unsigned int int_en  = READL(instance, &map_base[IND_REG_INTERRUPT_MASK]);
        unsigned int ind_count  = READL(instance, &map_base[IND_REG_INTERRUPT_COUNT]);
        PORTAL_PRINTF("(%d:fpga%d) about to receive messages int=%08x en=%08x qs=%08x cnt=%x\n", i, instance->fpga_number, int_src, int_en, queue_status, ind_count);
        instance->handler(instance, queue_status-1);
      }
    }
}

int nandsim_worker_thread_fn (void* arg) 
{
	while (1) {
		manual_event ();
		msleep (10);
		if (kthread_should_stop ()) {
			bdbm_msg ("nandsim_worker_thread_fn ends");
			break;
		}
	}
	complete (&worker_completion);
	return 0;
}

int nandsim_main_thread_fn (void* arg) 
{
	int srcAlloc;
	unsigned int *srcBuffer;
	unsigned int ref_srcAlloc;
	int rc = 0, i;
	pthread_t tid = 0;

	init_portal_internal(&intarr[0], IfcNames_DmaIndication, DmaIndicationWrapper_handleMessage);     // fpga1
	init_portal_internal(&intarr[1], IfcNames_NandSimIndication, NandSimIndicationWrapper_handleMessage); // fpga2
	init_portal_internal(&intarr[2], IfcNames_DmaConfig, DmaConfigProxy_handleMessage);         // fpga3
	init_portal_internal(&intarr[3], IfcNames_NandSimRequest, NandSimRequestProxy_handleMessage);    // fpga4

	sem_init(&test_sem, 0, 0);
	DmaManager_init(&priv, &intarr[2]);
	srcAlloc = portalAlloc(alloc_sz);
	if (rc){
		PORTAL_PRINTF("portal alloc failed rc=%d\n", rc);
		return rc;
	}

	PORTAL_PRINTF( "Main: creating exec thread\n");
	wake_up_process (kid_worker);
	srcBuffer = (unsigned int *)portalMmap(srcAlloc, alloc_sz);
#ifdef BSIM
	portalEnableInterrupts(&intarr[0]);
	portalEnableInterrupts(&intarr[1]);
	portalEnableInterrupts(&intarr[2]);
	portalEnableInterrupts(&intarr[3]);
#endif

	PORTAL_PRINTF( "Test 1: check for match\n");
	portalDCacheFlushInval(srcAlloc, alloc_sz, srcBuffer);
	for (i = 0; i < numWords; i++)
		srcBuffer[i] = i;
	PORTAL_PRINTF( "Main: before DmaManager_reference(%p)\n", srcAlloc);
	ref_srcAlloc = DmaManager_reference(&priv, srcAlloc);
	PORTAL_PRINTF( "Main::starting write - begin %08zx\n", numBytes);
	NandSimRequestProxy_startWrite (&intarr[3], ref_srcAlloc, 0, 0, numBytes, 1);
	PORTAL_PRINTF( "Main:: wait for semaphore\n");
	sem_wait(&test_sem);

	portalDCacheFlushInval(srcAlloc, alloc_sz, srcBuffer);
	for (i = 0; i < numWords; i++)
		srcBuffer[i] = 0;
	PORTAL_PRINTF( "Main::starting read %08zx\n", numBytes);
	NandSimRequestProxy_startRead (&intarr[3], ref_srcAlloc, 0, 0, numBytes, 1);
	sem_wait(&test_sem);
	PORTAL_PRINTF ("read: %u %u %u %u\n", srcBuffer[0], srcBuffer[1], srcBuffer[2], srcBuffer[3]);

	portalDCacheFlushInval(srcAlloc, alloc_sz, srcBuffer);
	PORTAL_PRINTF( "Main::starting erase %08zx\n", numBytes);
	NandSimRequestProxy_startErase (&intarr[3], 0, numBytes);
	sem_wait(&test_sem);

	portalDCacheFlushInval(srcAlloc, alloc_sz, srcBuffer);
	PORTAL_PRINTF( "Main::starting read %08zx\n", numBytes);
	NandSimRequestProxy_startRead (&intarr[3], ref_srcAlloc, 0, 0, numBytes, 1);
	sem_wait(&test_sem);
	PORTAL_PRINTF ("read: %u %u %u %u\n", srcBuffer[0], srcBuffer[1], srcBuffer[2], srcBuffer[3]);

	PORTAL_PRINTF( "Main: all done\n");

	bdbm_msg ("wait for a worker thread to finish");
	kthread_stop (kid_worker);
	wait_for_completion(&worker_completion);

	complete (&main_completion);
	main_program_finished = 1;

	return 0;
}

int bdbm_nandsim_thread_init (void)
{
	bdbm_msg ("bdbm_nandsim_thread_init minor %d\n", miscdev.minor);

	misc_register (&miscdev);
	if ((kid_main = kthread_create (nandsim_main_thread_fn, NULL, "nandsim_main_thread"))) {
		if ((kid_worker = kthread_create (nandsim_worker_thread_fn, NULL, "nandsim_worker_thread")) == NULL) {
			bdbm_error ("kthread_create failed");
			goto fail;
		}
		sema_init (&bsim_start, 0);
		wake_up_process (kid_main);
	} else
		goto fail;

	nandsim_init = 1;
	return 0;

fail:
	misc_deregister (&miscdev);
	return 1;
}

void bdbm_nandsim_thread_cleanup (void)
{
	if (!nandsim_init)
		return;

	bdbm_msg ("wait for a main thread to finish");
#ifdef BOARD_BSIM
	if (!bsim_relay_running) {
		printk("TestProgram::pa_exit terminate main program\n");
		main_program_finished = 1;
		printk("[%s:%d] - up\n", __FUNCTION__, __LINE__);
		up(&bsim_start); // in case host never starts
	}
	else
#endif
	wait_for_completion (&main_completion);

	misc_deregister (&miscdev);
}

