#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/delay.h>

#include <linux/pci.h>          /* pci device types, fns, etc. */
#include <linux/device.h>
#include <linux/cdev.h>
#include "../debug.h"

#include "../xbsv/cpp/portal_offsets.h"
#include "../xbsv/drivers/pcieportal/pcieportal.h"
#include "/home/chamdoo/simple_manual/vc707/jni/SimpleIndicationWrapperOffsets.h"
#include "/home/chamdoo/simple_manual/vc707/jni/SimpleRequestProxyOffsets.h"

#include "/home/chamdoo/memread_manual/vc707/jni/GeneratedTypes.h"


/* for pcieportal */
extern tBoard* get_pcie_portal_descriptor(void);

static int v1a = 42;
static int v2a = 2;
static int v2b = 4;

static tBoard* board = NULL;

volatile uint32_t* baseaddr; 


static struct task_struct *thread1 = NULL;



static int indication_handleMessage(volatile uint32_t* map_base, uint32_t channel)
{    
	uint32_t buf[1024];
	switch (channel) {
	case CHAN_NUM_SimpleIndicationWrapper_heard1: 
		{ 
			struct {
				uint32_t v:32;
			} payload;
			int i;
			for (i = (4/4)-1; i >= 0; i--)
				buf[i] = map_base[PORTAL_IND_FIFO(CHAN_NUM_SimpleIndicationWrapper_heard1)];
			i = 0;
			payload.v = (uint32_t)(((buf[i])&0xfffffffful));
			i++;
			bdbm_msg ("heard1(%d)", payload.v);
			break;
		}
	case CHAN_NUM_SimpleIndicationWrapper_heard2: 
		{ 
			struct {
				uint32_t a:32;
				uint32_t b:32;
			} payload;
			int i = 0;
			for (i = (8/4)-1; i >= 0; i--)
				buf[i] = map_base[PORTAL_IND_FIFO(CHAN_NUM_SimpleIndicationWrapper_heard2)];
			i = 0;
			payload.b = (uint32_t)(((buf[i])&0xfffffffful));
			i++;
			payload.a = (uint32_t)(((buf[i])&0xfffffffful));
			i++;
			bdbm_msg ("heard2(%d %d)", payload.a, payload.b);
			break;
		}
	default:
		bdbm_msg("SimpleIndicationWrapper::handleMessage: unknown channel 0x%x", channel);
		return 0;
	}
	return 0;
}

static int request_handleMessage(volatile uint32_t* map_base, uint32_t channel)
{    
	uint32_t buf[1024];
	switch (channel) {
	case CHAN_NUM_SimpleRequestProxyStatus_putFailed: 
		{ 
			struct {
				uint32_t v:32;
			} payload;
			const char* methodNameStrings[] = {"say1", "say2"};
			int i = 0;
			for (i = (4/4)-1; i >= 0; i--)
				buf[i] = map_base[PORTAL_IND_FIFO(CHAN_NUM_SimpleRequestProxyStatus_putFailed)];
			i = 0;
			payload.v = (uint32_t)(((buf[i])&0xfffffffful));
			i++;
			bdbm_msg("putFailed: %s", methodNameStrings[payload.v]);
			break;
		}
	default:
		bdbm_msg("SimpleRequestProxyStatus::handleMessage: unknown channel 0x%x", channel);
		return 0;
	}
	return 0;
}

static void manual_event(void)
{
	int i;
	for (i = 0; i < 2; i++) {
		volatile uint32_t *map_base = (volatile uint32_t *)(baseaddr + (i+1)*(1<<16)/4);
		uint32_t queue_status;
		while ((queue_status = map_base[IND_REG_QUEUE_STATUS])) {
			uint32_t int_src = map_base[IND_REG_INTERRUPT_FLAG];
			uint32_t int_en  = map_base[IND_REG_INTERRUPT_MASK];
			/*uint32_t ind_count  = map_base[IND_REG_INTERRUPT_COUNT];*/
			bdbm_msg("(%d) about to receive messages int=%08x en=%08x qs=%08x", i, int_src, int_en, queue_status);
			if (i == 1)
				indication_handleMessage(map_base, queue_status-1);
			else
				request_handleMessage(map_base, queue_status-1);
		}
	}
}

int bdbm_test_thread_simman_fn (void* arg) 
{
	bdbm_msg ("kthread starts");

	board = get_pcie_portal_descriptor ();

	bdbm_msg ("board = %p", board);
	bdbm_msg ("board->pci_dev = %p", board->pci_dev);
	bdbm_msg ("board->bar2io = %p", board->bar2io);

	baseaddr = board->bar2io;

	{
		volatile uint32_t* baseaddr_test = baseaddr + (0+1)*(1<<16)/4;
		bdbm_msg ("baseaddr[IND_REG_INTERRUPT_MASK] = %x", baseaddr_test[IND_REG_INTERRUPT_MASK]);
		baseaddr_test[IND_REG_INTERRUPT_MASK] = 1;
		bdbm_msg ("baseaddr[IND_REG_INTERRUPT_MASK] = %x", baseaddr_test[IND_REG_INTERRUPT_MASK]);
		baseaddr_test[IND_REG_INTERRUPT_MASK] = 0;
	}

	{
		volatile uint32_t* baseaddr_test = baseaddr + (1+1)*(1<<16)/4;
		bdbm_msg ("baseaddr[IND_REG_INTERRUPT_MASK] = %x", baseaddr_test[IND_REG_INTERRUPT_MASK]);
		baseaddr_test[IND_REG_INTERRUPT_MASK] = 1;
		bdbm_msg ("baseaddr[IND_REG_INTERRUPT_MASK] = %x", baseaddr_test[IND_REG_INTERRUPT_MASK]);
		baseaddr_test[IND_REG_INTERRUPT_MASK] = 0;
	}

	bdbm_msg ("Main::calling say1(%d)", v1a);
	{
		struct {
			uint32_t v:32;
		} payload;
		uint32_t buf[128];
		int i = 0;
		volatile uint32_t *map_base = (volatile uint32_t *)(baseaddr + 1*(1<<16)/4);

		payload.v = v1a;
		buf[i++] = payload.v;
		for (i = 4/4-1; i >= 0; i--)
			map_base[PORTAL_REQ_FIFO(CHAN_NUM_SimpleRequestProxy_say1)] = buf[i];
	};
	manual_event();

	bdbm_msg ("Main::calling say2(%d, %d)", v2a,v2b);
	{
		struct {
			uint32_t a:32;
			uint32_t b:32;
		} payload;
		uint32_t buf[128];
		int i = 0;
		volatile uint32_t *map_base = (volatile uint32_t *)(baseaddr + 1*(1<<16)/4);

		payload.a = v2a;
		payload.b = v2b;
		buf[i++] = payload.b;
		buf[i++] = payload.a;
		for (i = 8/4-1; i >= 0; i--)
			map_base[PORTAL_REQ_FIFO(CHAN_NUM_SimpleRequestProxy_say2)] = buf[i];
	};
	manual_event();

	/* wait for the module to finish */
	while (1) {	msleep(1); if (kthread_should_stop()) break; }

	return 0;
}

int bdbm_test_thread_fn (void* arg) 
{
	bdbm_msg ("kthread starts");

	while (1) {
		bdbm_msg ("kthread is running");

		mdelay(500); 
		msleep(1);

		if (kthread_should_stop()) {
			return 0;
		}
	}
	return 0;
}

int bdbm_test_thread_init (void) 
{
	if ((thread1 = kthread_create (bdbm_test_thread_simman_fn, NULL, "kthread"))) {
		wake_up_process (thread1);
	} else {
		bdbm_error ("kthread_create failed");
	}
	return 0;
}

void bdbm_test_thread_cleanup (void) 
{
	if (!kthread_stop (thread1)) {
		bdbm_msg ("kthread stops");
	}
}

