#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/blkdev.h>

#include <linux/pci.h>          /* pci device types, fns, etc. */
#include <linux/device.h>
#include <linux/cdev.h>
#include "../debug.h"

/* semaphores */
#include <linux/semaphore.h>

struct semaphore localmanager_confSem;
struct semaphore localmanager_mtSem;
struct semaphore localmanager_dbgSem;
struct semaphore test_sem;

/* for kernel thread */
static struct task_struct *kt = NULL;
static struct task_struct *tid = NULL;

/* for xbsv */
#include "pcieportal.h"
#include "portalmem.h"
#include "dmaManager.h"
#include "../xbsv/memread_manual/vc707/jni/GeneratedTypes.h"
#include "../xbsv/memread_manual/vc707/jni/DmaConfigProxy.cpp"
#include "../xbsv/memread_manual/vc707/jni/DmaIndicationWrapper.cpp"
#include "../xbsv/memread_manual/vc707/jni/MemreadIndicationWrapper.cpp"
#include "../xbsv/memread_manual/vc707/jni/MemreadRequestProxy.cpp"
#include "../xbsv/cpp/kdmaManager.cpp"

/* for pcieportal */
static tBoard* board = NULL;
volatile uint32_t* baseaddr; 

#define MAX_INDARRAY 4
typedef int (*INDFUNC)(PortalInternal *p, unsigned int channel);
static PortalInternal *intarr[MAX_INDARRAY];
static INDFUNC indfn[MAX_INDARRAY];
static int burstLen = 16;

#define numWords 0x1240000/4 // make sure to allocate at least one entry of each size
#define test_sz  numWords*sizeof(uint32_t)
static size_t alloc_sz = test_sz;
static DmaManagerPrivate priv;

void DmaConfigProxyStatusputFailed_cb (  struct PortalInternal *p, const uint32_t v )
{
        const char* methodNameStrings[] = {"sglist", "region", "addrRequest", "getStateDbg", "getMemoryTraffic"};
        printk( "putFailed: %s\n", methodNameStrings[v]);
}
void MemreadRequestProxyStatusputFailed_cb (  struct PortalInternal *p, const uint32_t v )
{
        const char* methodNameStrings[] = {"startRead"};
        printk( "putFailed: %s\n", methodNameStrings[v]);
}
void MemreadIndicationWrapperreadDone_cb (  struct PortalInternal *p, const uint32_t mismatchCount )
{
         printk( "Memread_readDone(mismatch = %x)\n", mismatchCount);
		 /*sem_post(&test_sem);*/
		 up (&test_sem);
}
void DmaIndicationWrapperconfigResp_cb (  struct PortalInternal *p, const uint32_t pointer, const uint64_t msg )
{
        //printk( "configResp: %x, %llu\n", pointer, msg);
        //printk( "configResp %d\n", pointer);
		/*sem_post(&priv.confSem);*/
		 up (&priv.confSem);
}
void DmaIndicationWrapperaddrResponse_cb (  struct PortalInternal *p, const uint64_t physAddr )
{
        printk( "DmaIndication_addrResponse(physAddr=%llu)\n", physAddr);
}
void DmaIndicationWrapperbadPointer_cb (  struct PortalInternal *p, const uint32_t pointer )
{
        printk( "DmaIndication_badPointer(pointer=%x)\n", pointer);
}
void DmaIndicationWrapperbadAddrTrans_cb (  struct PortalInternal *p, const uint32_t pointer, const uint64_t offset, const uint64_t barrier )
{
        printk( "DmaIndication_badAddrTrans(pointer=%x, offset=%llu barrier=%llu\n", pointer, offset, barrier);
}
void DmaIndicationWrapperbadPageSize_cb (  struct PortalInternal *p, const uint32_t pointer, const uint32_t sz )
{
        printk( "DmaIndication_badPageSize(pointer=%x, len=%x)\n", pointer, sz);
}
void DmaIndicationWrapperbadNumberEntries_cb (  struct PortalInternal *p, const uint32_t pointer, const uint32_t sz, const uint32_t idx )
{
        printk( "DmaIndication_badNumberEntries(pointer=%x, len=%x, idx=%x)\n", pointer, sz, idx);
}
void DmaIndicationWrapperbadAddr_cb (  struct PortalInternal *p, const uint32_t pointer, const uint64_t offset, const uint64_t physAddr )
{
        printk( "DmaIndication_badAddr(pointer=%x offset=%llu physAddr=%llu)\n", pointer, offset, physAddr);
}
void DmaIndicationWrapperreportStateDbg_cb (  struct PortalInternal *p, const DmaDbgRec rec )
{
        //printk( "reportStateDbg: {x:%08x y:%08x z:%08x w:%08x}\n", rec.x,rec.y,rec.z,rec.w);
        DmaDbgRec dbgRec = rec;
        printk( "dbgResp: %08x %08x %08x %08x\n", dbgRec.x, dbgRec.y, dbgRec.z, dbgRec.w);
		/*sem_post(&priv.dbgSem);*/
		up (&priv.dbgSem);
}
void DmaIndicationWrapperreportMemoryTraffic_cb (  struct PortalInternal *p, const uint64_t words )
{
        //printk( "reportMemoryTraffic: words=%llu\n", words);
        priv.mtCnt = words;
		/*sem_post(&priv.mtSem);*/
		up (&priv.mtSem);
}
void DmaIndicationWrappertagMismatch_cb (  struct PortalInternal *p, const ChannelType x, const uint32_t a, const uint32_t b )
{
        printk( "tagMismatch: %s %d %d\n", x==ChannelType_Read ? "Read" : "Write", a, b);
}

static void manual_event(void)
{
	int i;
    for (i = 0; i < MAX_INDARRAY; i++) {
      PortalInternal *instance = intarr[i];
      volatile uint32_t *map_base = instance->map_base;
      unsigned int queue_status;
      while ((queue_status= READL(instance, &map_base[IND_REG_QUEUE_STATUS]))) {
        uint32_t int_src = READL(instance, &map_base[IND_REG_INTERRUPT_FLAG]);
        uint32_t int_en  = READL(instance, &map_base[IND_REG_INTERRUPT_MASK]);
        uint32_t ind_count  = READL(instance, &map_base[IND_REG_INTERRUPT_COUNT]);
        printk( "(%d:%s) about to receive messages int=%08x en=%08x qs=%08x\n", i, "", int_src, int_en, queue_status);
        if (indfn[i])
            indfn[i](instance, queue_status-1);
      }
    }
}

int kthread_worker (void* arg) 
{
	printk ("kthread_worker starts\n");
    while (1) {
        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 10000;
		manual_event();
		msleep(10);

		if (kthread_should_stop()) {
			printk ("exit from kthread_worker\n");
			return 0;
		}
    }
	printk ("kthread_worker ends\n");
	return 0;
}

int bdbm_test_thread_memread_fn (void* arg) 
{
	PortalAlloc *srcAlloc;
	uint32_t *srcBuffer = NULL;
	uint32_t ref_srcAlloc;
	int rc;

	board = get_pcie_portal_descriptor ();
	baseaddr = board->bar2io;
	
	/* init semaphores */
	sema_init (&test_sem, 1);
  	down_interruptible (&test_sem);

	intarr[0] = kmalloc (sizeof (PortalInternal), GFP_KERNEL);
	intarr[1] = kmalloc (sizeof (PortalInternal), GFP_KERNEL);
	intarr[2] = kmalloc (sizeof (PortalInternal), GFP_KERNEL);
	intarr[3] = kmalloc (sizeof (PortalInternal), GFP_KERNEL);
	intarr[0]->map_base = baseaddr + (1)*(1<<16)/4;
	intarr[1]->map_base = baseaddr + (2)*(1<<16)/4;
	intarr[2]->map_base = baseaddr + (3)*(1<<16)/4;
	intarr[3]->map_base = baseaddr + (4)*(1<<16)/4;
  	indfn[0] = DmaIndicationWrapper_handleMessage;
  	indfn[1] = MemreadIndicationWrapper_handleMessage;
  	indfn[2] = DmaConfigProxyStatus_handleMessage;
  	indfn[3] = MemreadRequestProxyStatus_handleMessage;

	WRITEL(intarr[0], &intarr[0]->map_base[IND_REG_INTERRUPT_MASK], 0);
	WRITEL(intarr[0], &intarr[1]->map_base[IND_REG_INTERRUPT_MASK], 0);
	WRITEL(intarr[0], &intarr[2]->map_base[IND_REG_INTERRUPT_MASK], 0);
	WRITEL(intarr[0], &intarr[3]->map_base[IND_REG_INTERRUPT_MASK], 0);

	DmaManager_init(&priv, intarr[2]);
	rc = DmaManager_alloc(&priv, alloc_sz, &srcAlloc);
	if (rc){
		printk("portal alloc failed rc=%d\n", rc);
		return rc;
	}

	/* wake up kernel thread */
	wake_up_process (tid);

	/* initialize array */

	/* dcacheflush */

	ref_srcAlloc = DmaManager_reference(&priv, srcAlloc);
	printk( "Main: starting read %08x\n", numWords);
	MemreadRequestProxy_startRead (intarr[3] , ref_srcAlloc, numWords, burstLen, 1);
	down_interruptible (&test_sem);

	while (1) {	msleep(1); if (kthread_should_stop()) break; }

	return 0;
}
/*
 * EDN
 */

int bdbm_test_thread_fn (void* arg) 
{
	bdbm_msg ("kthread starts");

	while (1) {
		bdbm_msg ("kthread is running");

		mdelay(500); 
		msleep(1);

		if (kthread_should_stop()) {
			return 0;
		}
	}
	return 0;
}

int bdbm_test_memread_thread_init (void) 
{
	/*if ((kt = kthread_create (bdbm_test_thread_simman_fn, NULL, "kthread"))) {*/
	if ((kt = kthread_create (bdbm_test_thread_memread_fn, NULL, "kthread"))) {
		wake_up_process (kt);
		if ((tid = kthread_create (kthread_worker, NULL, "kthread_worker")) == NULL) {
			bdbm_error ("kthread_create failed");
			return 0;
		}
	} else {
		bdbm_error ("kthread_create failed");
	}
	return 0;
}

void bdbm_test_memread_thread_cleanup (void) 
{
	if (tid) {
		kthread_stop (tid);
		tid = NULL;
	}
	if (!kthread_stop (kt)) {
		bdbm_msg ("kthread stops");
	}
}

