#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kthread.h>
#include <linux/delay.h>

#include <linux/pci.h>          /* pci device types, fns, etc. */
#include <linux/device.h>
#include <linux/cdev.h>
#include "../debug.h"

/* for kernel thread */
static struct task_struct *kt = NULL;

/* for xbsv */
#include "pcieportal.h"
#include "../xbsv/simple_manual/vc707/jni/GeneratedTypes.h"
#include "../xbsv/simple_manual/vc707/jni/SimpleIndicationWrapper.cpp"
#include "../xbsv/simple_manual/vc707/jni/SimpleRequestProxy.cpp"

static int v1a = 42;
static int v2a = 2;
static int v2b = 4;
static tBoard* board = NULL;
static volatile uint32_t* baseaddr = NULL;

#define MAX_INDARRAY 2
typedef int (*INDFUNC)(PortalInternal *p, uint32_t channel);
static PortalInternal *intarr[MAX_INDARRAY];
static INDFUNC indfn[MAX_INDARRAY];

void SimpleRequestProxyStatusputFailed_cb (  struct PortalInternal *p, const uint32_t v )
{
    const char* methodNameStrings[] = {"say1", "say2"};
    printk("putFailed: %s\n", methodNameStrings[v]);
}
void SimpleIndicationWrapperheard1_cb (  struct PortalInternal *p, const uint32_t v )
{
    printk("heard1(%d)\n", v);
}
void SimpleIndicationWrapperheard2_cb (  struct PortalInternal *p, const uint32_t a, const uint32_t b )
{
    printk("heard2(%d %d)\n", a, b);
}

static void manual_event(void)
{
	int i;
    for (i = 0; i < 2; i++) {
      PortalInternal *instance = intarr[i];
      volatile uint32_t *map_base = instance->map_base;
      uint32_t queue_status;
      while ((queue_status= READL(instance, &map_base[IND_REG_QUEUE_STATUS]))) {
        uint32_t int_src = READL(instance, &map_base[IND_REG_INTERRUPT_FLAG]);
        uint32_t int_en  = READL(instance, &map_base[IND_REG_INTERRUPT_MASK]);
        uint32_t ind_count  = READL(instance, &map_base[IND_REG_INTERRUPT_COUNT]);
		/*printk("(%d:%s) about to receive messages int=%08x en=%08x qs=%08x\n", i, instance->name, int_src, int_en, queue_status);*/
		printk("(%d:%s) about to receive messages int=%08x en=%08x qs=%08x\n", i, "", int_src, int_en, queue_status);
        if (i == 0)
            SimpleRequestProxyStatus_handleMessage(instance, queue_status-1);
        else
            SimpleIndicationWrapper_handleMessage(instance, queue_status-1);
      }
    }
}

int bdbm_test_thread_simman_fn (void* arg) 
{
	board = get_pcie_portal_descriptor ();
	bdbm_msg ("%p %p", board, board->bar2io);
	
	baseaddr = board->bar2io;

	intarr[0] = kmalloc (sizeof (PortalInternal), GFP_KERNEL);
	intarr[1] = kmalloc (sizeof (PortalInternal), GFP_KERNEL);
	intarr[0]->map_base = baseaddr + (1)*(1<<16)/4;
	intarr[1]->map_base = baseaddr + (2)*(1<<16)/4;
	indfn[0] = SimpleRequestProxyStatus_handleMessage;
	indfn[1] = SimpleIndicationWrapper_handleMessage;

	bdbm_msg ("%p %p %p %p", intarr[0], intarr[1], indfn[0], indfn[1]);

	WRITEL(intarr[0], &intarr[0]->map_base[IND_REG_INTERRUPT_MASK], 0);
	WRITEL(intarr[0], &intarr[1]->map_base[IND_REG_INTERRUPT_MASK], 0);
	printk("Main::calling say1(%d)\n", v1a);
	SimpleRequestProxy_say1 (intarr[0] , v1a);
	manual_event();

	printk("Main::calling say2(%d, %d)\n", v2a,v2b);
	SimpleRequestProxy_say2 (intarr[0], v2a, v2b);
	manual_event();


	while (1) {	msleep(1); if (kthread_should_stop()) break; }

	return 0;
}

int bdbm_test_thread_fn (void* arg) 
{
	bdbm_msg ("kthread starts");

	while (1) {
		bdbm_msg ("kthread is running");

		mdelay(500); 
		msleep(1);

		if (kthread_should_stop()) {
			return 0;
		}
	}
	return 0;
}

int bdbm_test_thread_init (void) 
{
	if ((kt = kthread_create (bdbm_test_thread_simman_fn, NULL, "kthread"))) {
		wake_up_process (kt);
	} else {
		bdbm_error ("kthread_create failed");
	}
	return 0;
}

void bdbm_test_thread_cleanup (void) 
{
	if (!kthread_stop (kt)) {
		bdbm_msg ("kthread stops");
	}
}

