#include <linux/module.h>
#include <linux/ktime.h>
#include <linux/time.h>
#include <linux/slab.h>

#include "bdbm_drv.h"
#include "platform.h"
#include "debug.h"
#include "time.h"


#ifdef USE_KTIMER
  ktime_t ktime_get (void);
#endif


static uint32_t _time_startup_timestamp = 0;


uint32_t time_get_timestamp_in_us (void)
{
	struct timeval tv;
	do_gettimeofday(&tv);
	return tv.tv_sec * 1000000 + tv.tv_usec;
}

uint32_t time_get_timestamp_in_sec (void)
{
	struct timeval tv;
	do_gettimeofday(&tv);
	return tv.tv_sec - _time_startup_timestamp;
}

void time_init (void)
{
	_time_startup_timestamp = time_get_timestamp_in_sec ();
}


/* stopwatch functions */
/* taken from http://www.gnu.org/software/libc/manual/html_node/Elapsed-Time.html */
int timeval_subtract (
	struct timeval *result, 
	struct timeval *x, 
	struct timeval *y)
{
	if (x->tv_usec < y->tv_usec) {
		int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		y->tv_usec -= 1000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000) {
		int nsec = (x->tv_usec - y->tv_usec) / 1000000;
		y->tv_usec += 1000000 * nsec;
		y->tv_sec -= nsec;
	}

	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_usec = x->tv_usec - y->tv_usec;

	return x->tv_sec < y->tv_sec;
}

void bdbm_stopwatch_start (struct bdbm_stopwatch* sw)
{
	if (sw) {
#ifdef USE_KTIMER
		sw->start = ktime_get ();
#else
		do_gettimeofday(&sw->start);
#endif
	}
}

int64_t bdbm_stopwatch_get_elapsed_time_ms (struct bdbm_stopwatch* sw)
{
	if (sw) {
#ifdef USE_KTIMER
		ktime_t end = ktime_get ();
		return ktime_to_ms (ktime_sub (end, sw->start));
#else
		struct timeval diff, end;
		do_gettimeofday (&end);
		if (timeval_subtract (&diff, &end, &sw->start) == 0) {
			return (diff.tv_sec * 1000000 + diff.tv_usec)/1000;
		}
#endif
	}
	return 0;
}

int64_t bdbm_stopwatch_get_elapsed_time_us (struct bdbm_stopwatch* sw)
{
	if (sw) {
#ifdef USE_KTIMER
		ktime_t end = ktime_get ();
		return ktime_to_us (ktime_sub (end, sw->start));
#else
		struct timeval diff, end;
		do_gettimeofday (&end);
		if (timeval_subtract (&diff, &end, &sw->start) == 0) {
			return diff.tv_sec * 1000000 + diff.tv_usec;
		}
#endif
	}
	return 0;
}

